package dev.pfaff.altargui.widgets;

import dev.pfaff.altargui.*;
import dev.pfaff.altargui.element.Painter;
import dev.pfaff.altargui.internal.util.StencilLayoutContext;
import dev.pfaff.altargui.render.objects.OutlineRenderObject;
import dev.pfaff.altargui.types.Color;
import org.jetbrains.annotations.NotNull;

/**
 * @param color the RGBA colorRgba of the widget
 */
public record BorderWidget(int color, float thickness, Widget child) implements SingleChildWidget {
	public static final BorderWidget TRANSPARENT = new BorderWidget(Color.TRANSPARENT, 0f, Widgets.EMPTY);

	/**
	 * @deprecated Use {@link #BorderWidget(Color, float, Widget)} instead.
	 */
	@SuppressWarnings("DeprecatedIsStillUsed")
	@Deprecated
	public BorderWidget {
	}

	public BorderWidget(Color color, float thickness, Widget child) {
		this(color.rgba(), thickness, child);
	}

	/**
	 * @param color ARGB
	 */
	public static BorderWidget solid(Color color, float thickness, Widget child) {
		if (thickness == 0f && (child == null || child == Widgets.EMPTY)) return TRANSPARENT;
		return new BorderWidget(color, thickness, child);
	}

	@Override
	public @NotNull WidgetInstance<BorderWidget> instantiate(WidgetInstantiationArgs args) {
		return new State(args);
	}

	public BorderWidget withColor(Color color) {
		return new BorderWidget(color, thickness, child);
	}

	public BorderWidget withThickness(float thickness) {
		return new BorderWidget(color, thickness, child);
	}

	private static final class State extends CustomLayoutState<BorderWidget> implements Painter {
		private final OutlineRenderObject outline = new OutlineRenderObject();
		private int color;
		private float thickness;

		private State(WidgetInstantiationArgs args) {
			super(args);
		}

		@Override
		public void configure(StateContext stateContext, BorderWidget widget, boolean init) {
			super.configure(stateContext, widget, init);
			if (widget.color != this.color) {
				this.outline.updateColor(Color.fromRgba(this.color = widget.color));
			}
			if (widget.thickness != this.thickness) {
				this.thickness = widget.thickness;
				this.markNeedsLayout();
			}
		}

		@Override
		protected void computeStencilLayout(StateContext context) {
			computeFillLayout(this);
			StencilLayoutContext.updateMountAndLayout(context, this, this.getChildElement(), (e, ctx) -> {
				var thickness = this.thickness;
				float thickness2 = thickness * 2;
				e.setAvailableSize(this.allocatedWidth() - thickness2,
						this.allocatedHeight() - thickness2);
				e.stencilLayout(ctx);
				float width = e.computedWidth();
				float height = e.computedHeight();
				this.setComputedSize(width + thickness2, height + thickness2);
			});
		}

		@Override
		protected void computeTraceLayout() {
			var e = getChildElement();
			e.setPosition(this.allocatedX() + thickness, this.allocatedY() + thickness);
			e.traceLayout();
			float x = e.allocatedX();
			float y = e.allocatedY();
			float width = e.computedWidth();
			float height = e.computedHeight();
			this.outline.updatePositionSizeThickness(x, y, width, height, thickness);
		}

		@Override
		public void paint(long time, Canvas canvas) {
			var color = this.color;

			if (Color.fromRgba(color).alpha() == 0) return;

			canvas.upload(outline);
		}
	}
}
